@@ -19,6 +19,7 @@
 #    4-Jan-2019  jdw add optional arguments to getPath(...,prefixName=None, prefixSectionName=None)
 #                    add methods getDefaultSectionName(), replaceSectionName(), and getSectionNameReplacement()
 #   10-Mar-2019  jdw add method getEnvValue() to dereference config option as an environmental variable
+#    3-Feb-2020  jdw add __processAppendedSections() to handle nested configuration sections
 ##
 """
  Manage simple configuration options.
@@ -35,11 +36,14 @@ import copy
 import logging
 import os
 import sys
+import tempfile
 
 import ruamel.yaml
 from nacl.encoding import HexEncoder
 from nacl.secret import SecretBox
 
+from rcsb.utils.io.FileUtil import FileUtil
+
 try:
     from configparser import ConfigParser as cp
 except ImportError:
@@ -50,7 +54,18 @@ logger = logging.getLogger(__name__)
 
 
 class ConfigUtil(object):
-    def __init__(self, configPath=None, defaultSectionName="DEFAULT", fallbackEnvPath=None, mockTopPath=None, configFormat=None, **kwargs):
+    def __init__(
+        self,
+        configPath=None,
+        defaultSectionName="DEFAULT",
+        fallbackEnvPath=None,
+        mockTopPath=None,
+        configFormat=None,
+        cachePath=None,
+        useCache=False,
+        appendConfigOption="CONFIG_APPEND_LOCATOR_PATHS",
+        **kwargs
+    ):
         """Manage simple configuration options stored in INI (Python configparser-style) or YAML configuration files.
 
 
@@ -60,6 +75,9 @@ class ConfigUtil(object):
             fallbackEnvPath (str, optional): Environmental variable holding configuration file path
             mockTopPath (str, optional): Mockpath is prepended to path configuration options if it specified (default=None)
             configFormat (str, optional): Configuration file format (e.g. ini or yaml default=ini)
+            cachePath (str,optional): top cache path for remote configuration assets
+            useCache (bool, optional): use cached configuration assets
+            appendConfigOption (str, optional): option name containing a list of appendable configuration files (yaml only)
             **kwargs: importEnvironment(bool) imports environment as default values for ini/configparser format files
 
 
@@ -72,6 +90,10 @@ class ConfigUtil(object):
         #  Mockpath is prepended to path configuration options if it specified
         self.__mockTopPath = mockTopPath
         #
+        # Top cache path for remote configuration assets - if no cache information is provided then use system temp area.
+        cachePath = cachePath if cachePath else tempfile.mkdtemp(prefix="config-util", suffix="-cache")
+        appendConfigOption = appendConfigOption if appendConfigOption else None
+        #
         # This is the internal container for configuration data from all sources
         self.__cD = {}
         #
@@ -83,9 +105,48 @@ class ConfigUtil(object):
         logger.debug("Using config path %s format %s", self.__myConfigPath, self.__configFormat)
         if self.__myConfigPath:
             self.__configFormat, self.__cD = self.__updateConfig(self.__myConfigPath, self.__configFormat, **kwargs)
+            self.__processAppendedSections(appendConfigOption, cachePath, useCache)
             if not self.__cD:
                 logger.warning("No configuration information imported - configuration path is %s (%s)", self.__myConfigPath, configFormat)
 
+    def __processAppendedSections(self, appendConfigOption, cachePath, useCache=True):
+        """Fetch and append configuration assets assigned to input configuration option.
+
+        Args:
+            appendConfigOption (str): reserved configuration option to hold a list of configuration asset locators
+            cachePath (str): path to store cached copies configuration assets
+            useCache (bool, optional): use existing cached configuration assets. Defaults to True.
+
+        Returns:
+            bool: True for success of False otherwise
+        """
+        try:
+            ret = True
+            appendLocL = self.getList(appendConfigOption, sectionName=self.__defaultSectionName)
+            logger.debug("appendLocL is %r", appendLocL)
+            if appendLocL:
+                cP = os.path.join(cachePath, "config")
+                fU = FileUtil(workPath=cP)
+                logger.debug("Fetching append sections from %r", appendLocL)
+                for appendLoc in appendLocL:
+                    fn = fU.getFileName(appendLoc)
+                    fp = os.path.join(cP, fn)
+                    okF = True
+                    if not (useCache and fU.exists(fp)):
+                        # get a fresh copy from source
+                        okF = fU.get(appendLoc, fp)
+                        logger.debug("Fetched %r to %r", appendLoc, fp)
+                    ok = self.appendConfig(fp)
+                    ret = ret and ok and okF
+        except Exception as e:
+            logger.exception("Failing for option %r cachePath %r with %s", appendConfigOption, cachePath, str(e))
+            ret = False
+        #
+        if not ret:
+            logger.error("Fetching appended sections failing %r", appendLocL)
+
+        return ret
+
     def getConfigPath(self):
         return self.__myConfigPath
 
@@ -234,6 +295,43 @@ class ConfigUtil(object):
         return ok
 
     def get(self, name, default=None, sectionName=None, tokenName="CONFIG_SUPPORT_TOKEN"):
+        """Return configuration value of input configuration option. Option names beginning with
+           leading underscore are treated as encrypted secrets. If an encrypted option is
+           not found in the section this method will fallback to the value of the unqualified
+           option.
+
+        Args:
+            name (str): configuration option name
+            default (str, optional): default value returned if no configuration option is provided
+            sectionName (str, optional): configuration section name, a simple key (default = defaultSectionName from object)
+            tokenName (str,optional): configuration option holding name of environmental variable
+                                        storing security key.
+        Returns:
+            str: configuration option value
+
+        """
+        logMissing = False
+        ok = False
+        mySection = sectionName if sectionName else self.__defaultSectionName
+        mySection = self.getSectionNameReplacement(mySection)
+        try:
+            if "." in name:
+                ok = self.__getKeyExists(self.__cD[mySection], name)
+            else:
+                ok = name in self.__cD[mySection]
+        except Exception:
+            ok = False
+        #
+        if ok:
+            return self.__get(name, default=default, sectionName=sectionName, tokenName=tokenName)
+        elif name.startswith("_"):
+            return self.__get(name[1:], default=default, sectionName=sectionName, tokenName=tokenName)
+        else:
+            if logMissing:
+                logger.debug("Missing config option %r (%r) assigned default value %r", name, mySection, default)
+            return default
+
+    def __get(self, name, default=None, sectionName=None, tokenName="CONFIG_SUPPORT_TOKEN"):
         """Return configuration value of input configuration option. Option names beginning with
            leading underscore are treated as encrypted secrets.
 
@@ -260,6 +358,7 @@ class ConfigUtil(object):
             val = str(val) if self.__configFormat == "ini" else val
             if name.startswith("_") and isinstance(val, str):
                 val = self.__getSecretValue(name, val, mySection, tokenName)
+
         except Exception as e:
             if logMissing:
                 logger.debug("Missing config option %r (%r) assigned default value %r (%s)", name, mySection, default, str(e))
@@ -304,11 +403,14 @@ class ConfigUtil(object):
     def __getSecretValue(self, name, val, sectionName, tokenName):
         try:
             hexKey = self.getEnvValue(tokenName, sectionName=sectionName)
-            if val and hexKey:
-                val = self.__decryptMessage(val, hexKey)
+            if not hexKey:
+                logger.error("Empty key for token %r processing %r and %r", tokenName, name, val)
+            elif len(hexKey) < 32:
+                logger.error("Bad key (%d) for token %r processing %r and %r", len(hexKey), tokenName, name, val)
+            val = self.__decryptMessage(val, hexKey)
             hexKey = None
         except Exception as e:
-            logger.error("Failing processing %s secret value with %s", name, str(e))
+            logger.debug("Failing processing %s using %r secret value with %s", name, tokenName, str(e))
         return val
 
     def getSecret(self, name, default=None, sectionName=None, tokenName="CONFIG_SUPPORT_TOKEN"):
@@ -349,7 +451,7 @@ class ConfigUtil(object):
             varName = self.get(name, default=None, sectionName=sectionName)
             val = os.environ.get(varName, default)
         except Exception as e:
-            logger.debug("Failed processing environmental variable config option %r (%r) assigned default value %r (%s)", name, sectionName, default, str(e))
+            logger.error("Failed processing environmental variable config option %r (%r) assigned default value %r (%s)", name, sectionName, default, str(e))
         #
         return val
 
@@ -357,10 +459,12 @@ class ConfigUtil(object):
         vL = default if default is not None else []
         try:
             val = self.get(name, default=default, sectionName=sectionName)
-            if isinstance(val, (list, set, tuple)):
-                vL = list(val)
-            else:
-                vL = str(val).split(delimiter)
+            logger.debug("name %r sectionName %r val %r", name, sectionName, val)
+            if val:
+                if isinstance(val, (list, set, tuple)):
+                    vL = list(val)
+                else:
+                    vL = str(val).split(delimiter)
         except Exception as e:
             logger.debug("Missing config option list %r (%r) assigned default value %r (%s)", name, sectionName, default, str(e))
         #
@@ -568,6 +672,21 @@ class ConfigUtil(object):
 
         return None
 
+    def __getKeyExists(self, dct, keyName):
+        """  Return the key expressed in dot notation is in the input dictionary object (nested).
+        """
+        try:
+            kys = keyName.split(".")
+            for key in kys:
+                try:
+                    dct = dct[key]
+                except KeyError:
+                    return False
+            return True
+        except Exception as e:
+            logger.exception("Failing for key %r with %s", keyName, str(e))
+        return False
+
     def dump(self):
         for section in self.__cD:
             logger.info("Configuration section: %s", section)
@@ -592,6 +711,6 @@ class ConfigUtil(object):
             logger.debug("type %r text %r", type(dcrMsg), dcrMsg)
             txt = dcrMsg.decode("utf-8")
         except Exception as e:
-            logger.exception("Failing with %s", str(e))
+            logger.debug("Failing with %s", str(e))
 
         return txt
